#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include "ccpudef.h"

int inputLineNumber=0;

void parseError(std::string message)
{
	std::cout << "line " << inputLineNumber << ": " << message << std::endl;
	exit(-1);
}

unsigned short baseAddress=0;

std::vector<unsigned short> outputBuffer;

struct Label {
	std::string name;
	int lineNumber;
	unsigned short address;
};

std::vector<Label> labels;
std::vector<Label> labelReferences;

void registerLabel(std::string name)
{
	Label label;
	label.name=name;
	label.address=baseAddress+outputBuffer.size();

	labels.push_back(label);
}

unsigned short getLabelAddress(std::string name)
{
	for(unsigned int i=0;i<labels.size();i++) {
		if(labels[i].name==name) {
			return labels[i].address;
		}
	}
	
	parseError("undefined label " + name);
}

void addLabelReference(std::string name, unsigned short bufferOffset)
{
	Label label;
	label.name=name;
	label.lineNumber=inputLineNumber;
	label.address=bufferOffset;

	labelReferences.push_back(label);
}

void resolveLabelReferences()
{
	int origLineNumber=inputLineNumber;
	for(unsigned int i=0;i<labelReferences.size();i++) {
		inputLineNumber=labelReferences[i].lineNumber;
		unsigned short address=getLabelAddress(labelReferences[i].name);
		outputBuffer[labelReferences[i].address]=address;
	}
	inputLineNumber=origLineNumber;
}

bool isLabel(std::string operand)
{
	if(operand!="" && operand[0]=='$') {
		return true;
	}
	return false;
}

void parseLabel(std::string operand, unsigned short instructionOffset)
{
	addLabelReference(operand.substr(1,operand.length()-1),outputBuffer.size()+instructionOffset);
}

bool isRegister(std::string operand)
{
	if(operand=="A" || operand=="B" || operand=="C" || operand=="IP" || operand=="SP" || operand=="FLAGS") {
		return true;
	}
	return false;
}

CCPUOperandCode parseRegister(std::string operand)
{
	if(operand=="A") {
		//std::cout << "A" << std::endl;
		return RegisterA;
	} else if(operand=="B") {
		//std::cout << "B" << std::endl;
		return RegisterB;
	} else if(operand=="C") {
		//std::cout << "C" << std::endl;
		return RegisterC;
	} else if(operand=="IP") {
		//std::cout << "IP" << std::endl;
		return RegisterIP;
	} else if(operand=="SP") {
		//std::cout << "SP" << std::endl;
		return RegisterSP;
	} else if(operand=="FLAGS") {
		//std::cout << "FLAGS" << std::endl;
		return RegisterFLAGS;
	} else {
		parseError("unknown register " + operand);
	}
}

bool isBinary(std::string operand)
{
	if(operand.length()>2 && operand.substr(0,2)=="0b") {
		return true;
	}
	return false;
}

int parseBinary(std::string value)
{
	int result=0;

	for(int i=0;i<value.length();i++) {
		result*=2;
		switch(value[i]) {
		case '0': break;
		case '1': result+=1; break;
		}
	}

	return result;
}

bool isHex(std::string operand)
{
	if(operand.length()>2 && operand.substr(0,2)=="0x") {
		return true;
	}
	return false;
}

int parseHex(std::string value)
{
	int result=0;

	for(int i=0;i<value.length();i++) {
		result*=16;
		switch(value[i]) {
		case '0': break;
		case '1': result+=1; break;
		case '2': result+=2; break;
		case '3': result+=3; break;
		case '4': result+=4; break;
		case '5': result+=5; break;
		case '6': result+=6; break;
		case '7': result+=7; break;
		case '8': result+=8; break;
		case '9': result+=9; break;
		case 'A': result+=10; break;
		case 'B': result+=11; break;
		case 'C': result+=12; break;
		case 'D': result+=13; break;
		case 'E': result+=14; break;
		case 'F': result+=15; break;
		}
	}

	return result;
}

unsigned short parseNumeric(std::string string, unsigned short instructionOffset)
{
	int value;

	if(isLabel(string)) {
		parseLabel(string,instructionOffset);
		value=0xFFFF;
	} else if(isBinary(string)) {
		value=parseBinary(string.substr(2,string.length()-2));
	} else if(isHex(string)) {
		value=parseHex(string.substr(2,string.length()-2));
	} else {
		value=atoi(string.c_str());
	}

	return value;
}

CCPUOperandCode parseOperand(std::string operand, unsigned short &value, unsigned short &instructionOffset)
{
	if(operand.empty()) {
		parseError("missing operand");
	}

	if(operand[0]=='[') {
		// memory addressing
		std::string interior=operand.substr(1,operand.length()-2);

		int plusPos=interior.find("+");
		if(plusPos!=std::string::npos) {
//			std::cout << "base/offset addr" << std::endl;
			// base/offset addressing
			std::string base=interior.substr(0,plusPos);
			std::string offset=interior.substr(plusPos+1,interior.length()-plusPos-1);

			if(isRegister(base)) {
				if(parseRegister(base)!=RegisterB || parseRegister(offset)!=RegisterA) {
					parseError("base must be either B or numeric value when using base/offset addressing");
				}
				return MemoryBPlusA;
			}
			if(!isRegister(offset)) {
				parseError("offset cannot be a constant if base is");
			}
			CCPUOperandCode reg=parseRegister(offset);
			value=parseNumeric(base,instructionOffset);
			instructionOffset++;
			// not all registers are valid here, handle it while translating code
			switch(reg) {
			case RegisterA: return MemoryConstantPlusA;
			case RegisterIP: return MemoryConstantPlusIP;
			case RegisterSP: return MemoryConstantPlusSP;
			default:
				parseError("reigster not allowed as offset in base/offset addressing");
			}
		} else {
			// direct addressing
//			std::cout << "direct addr" << std::endl;
			if(isRegister(interior)) {
				CCPUOperandCode reg=parseRegister(interior);
				// not all registers are valid here, handle it while translating code
				switch(reg) {
				case RegisterA: return MemoryA;
				case RegisterB: return MemoryB;
				case RegisterC: return MemoryC;
				case RegisterSP: return MemorySP;
				default:
					parseError("register not allowed for direct memory addressing");
					break;
				}
			} else {
				value=parseNumeric(interior,instructionOffset);
				instructionOffset++;
				return MemoryConstant;
			}
		}
	} else {
		if(isRegister(operand)) {
			return parseRegister(operand);
		} else {
			value=parseNumeric(operand,instructionOffset);
			instructionOffset++;
			return Constant;
		}
	}
}

void emitConditional(CCPUOperandCode operandCode, unsigned short value)
{
	if(operandCode==Constant || operandCode==MemoryConstant || operandCode==MemoryConstantPlusA || operandCode==MemoryConstantPlusIP || operandCode==MemoryConstantPlusSP) {
		outputBuffer.push_back(value);
	}
}

void emit(CCPUInstruction instruction, CCPUOperandCode operandCode1, CCPUOperandCode operandCode2, CCPUOperandCode operandCode3, unsigned short value1, unsigned short value2, unsigned short value3)
{
	unsigned short opcode;

	opcode=(instruction<<12) + (operandCode1<<8) + (operandCode2<<4) + operandCode3;

	outputBuffer.push_back(opcode);

	emitConditional(operandCode1,value1);
	emitConditional(operandCode2,value2);
	emitConditional(operandCode3,value3);
}

void emitWords(std::string buffer)
{
	std::istringstream line(buffer);
	std::string temp;
	line >> temp;

	while(!line.eof())  {
		std::string valueStr;
		line >> valueStr;
		unsigned short value;
		value=parseNumeric(valueStr,0);
		outputBuffer.push_back(value);
	}
}

void emitString(std::string line)
{
	int quotationMarkPos=line.find("\"");
	if(quotationMarkPos==std::string::npos) {
		parseError("expected string literal start not found");
	}
	line=line.substr(quotationMarkPos+1,line.length()-quotationMarkPos-1);

	quotationMarkPos=line.find("\"");
	if(quotationMarkPos==std::string::npos) {
		parseError("expected string literal end not found");
	}
	line=line.substr(0,quotationMarkPos);

	for(int i=0;i<line.length();i++) {
		outputBuffer.push_back(line[i]);
	}
}

void emitFill(unsigned short value, unsigned short length)
{
//	std::cout << "fill length " << length << std::endl;
	for(unsigned short i=0;i<length;i++) {
		outputBuffer.push_back(value);
	}
}

void emitFillTo(unsigned short value, unsigned short offset)
{
//	std::cout << "fill to " << offset << std::endl;
	if(offset<outputBuffer.size()) {
		parseError("output already reaches beyond fill target address");
	}

	emitFill(value,offset-outputBuffer.size());
}

int main(int argc, char *argv[])
{
	std::ifstream input(argv[1]);
	std::ofstream output(argv[2],std::ios::binary);

	if(!input) {
		std::cout << "unable to open input file"  << std::endl;
		return 1;
	}

	if(!output) {
		std::cout << "unable to open output file" << std::endl;
		return 1;
	}

	while(!input.eof()) {
		std::string buffer;
		inputLineNumber++;
		char c;
		do {
			c=input.get();
			if(!input.eof() && c!='\n' && c!='\r') {
				buffer+=c;
			}
		} while(c!='\n' && !input.eof());
		
//		std::cout << buffer << std::endl;

		while(buffer.length()!=0 && isspace(buffer[0])) {
			buffer=buffer.substr(1,buffer.length()-1);
		}
		for(int i=0;i<buffer.length();i++) {
			if(buffer[i]=='#') {
				buffer=buffer.substr(0,i);
				break;
			}
		}

		if(buffer.length()==0) {
			continue;
		}

		std::istringstream line(buffer);

//		std::cout << line.str() << std::endl;

		// TODO trim comments, skip empty lines

		std::string mnemonic;
		line >> mnemonic;
//		std::cout << "mnemonic " << mnemonic << std::endl;

		std::string operand1, operand2, operand3;
		unsigned short operand1Value=0, operand2Value=0, operand3Value=0;

		CCPUInstruction instruction;
		CCPUOperandCode operand1Code=RegisterA, operand2Code=RegisterA, operand3Code=RegisterA;
		unsigned short instructionOffset=1;

		bool emitInstruction=true;

		if(mnemonic=="nop") {
			instruction=InstNOP;
		} else if(mnemonic=="add") {
			instruction=InstADD;
			line >> operand1 >> operand2 >> operand3;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="sub") {
			instruction=InstSUB;
			line >> operand1 >> operand2 >> operand3;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="mul") {
			instruction=InstMUL;
			line >> operand1 >> operand2 >> operand3;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="div") {
			instruction=InstDIV;
			line >> operand1 >> operand2 >> operand3;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="jmp") {
			instruction=InstJUMP;
			line >> operand3;
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="call") {
			instruction=InstCALL;
			line >> operand3;
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="move") {
			instruction=InstMOVE;
			line >> operand1 >> operand3;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="te") {
			instruction=InstTE;
			line >> operand1 >> operand2;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
		} else if(mnemonic=="tl") {
			instruction=InstTL;
			line >> operand1 >> operand2;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
		} else if(mnemonic=="tle") {
			instruction=InstTLE;
			line >> operand1 >> operand2;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
		} else if(mnemonic=="in") {
			instruction=InstIN;
			line >> operand1 >> operand2;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
		} else if(mnemonic=="out") {
			instruction=InstOUT;
			line >> operand1 >> operand2;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
			operand2Code=parseOperand(operand2,operand2Value,instructionOffset);
		} else if(mnemonic=="push") {
			instruction=InstPUSH;
			line >> operand1;
			operand1Code=parseOperand(operand1,operand1Value,instructionOffset);
		} else if(mnemonic=="pop") {
			instruction=InstPOP;
			line >> operand3;
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic=="cj") {
			instruction=InstCJ;
			line >> operand3;
			operand3Code=parseOperand(operand3,operand3Value,instructionOffset);
		} else if(mnemonic==".baseaddr") {
			emitInstruction=false;
			std::string value;
			line >> value;
			baseAddress=parseNumeric(value,0);
		} else if(mnemonic==".word") {
			emitInstruction=false;
			emitWords(buffer);
		} else if(mnemonic==".string") {
			emitInstruction=false;
			emitString(buffer);
		} else if(mnemonic==".fill") {
			emitInstruction=false;
			std::string value1, value2;
			line >> value1 >> value2;
			emitFill(parseNumeric(value1,0),parseNumeric(value2,0));
		} else if(mnemonic==".fillto") {
			emitInstruction=false;
			std::string value1, value2;
			line >> value1 >> value2;
			emitFillTo(parseNumeric(value1,0),parseNumeric(value2,0));
		} else if(mnemonic[0]=='$') {
			emitInstruction=false;
			registerLabel(mnemonic.substr(1,mnemonic.size()-1));
		} else {
			parseError("unknown mnemonic " + mnemonic);
		}

		if(emitInstruction) {
			emit(instruction,operand1Code,operand2Code,operand3Code,operand1Value,operand2Value,operand3Value);
		}
	}

	resolveLabelReferences();

	for(int i=0;i<outputBuffer.size();i++) {
		unsigned char low=outputBuffer[i] & 0xFF;
		unsigned char high=outputBuffer[i] >> 8;
		output.write((char*)&high,1);
		output.write((char*)&low,1);
	}

	return 0;
}
