#include <fstream>
#include <iostream>
#include <string>

#include "SDL.h"

#include "ccpudef.h"

#define LITTLE_ENDIAN_HOST

void fixBufferEndianess(unsigned short *buffer, int length)
{
#ifdef LITTLE_ENDIAN_HOST
	unsigned char *byteBuffer=reinterpret_cast<unsigned char*>(buffer);
	for(unsigned int i=0;i<length;i++) {
		unsigned char temp;
		temp=byteBuffer[2*i];
		byteBuffer[2*i]=byteBuffer[2*i+1];
		byteBuffer[2*i+1]=temp;
	}
#endif
}

class CCPUDevice {
public:
	virtual unsigned short in(unsigned char port) = 0;
	virtual void out(unsigned char port, unsigned short value) = 0;
};

class CCPUStorage : public CCPUDevice {
public:
	static const int blockSize=512;
	static const int numBlocks=0x0400;

	CCPUStorage(std::string imageFile, unsigned char commandPort, unsigned char dataPort);

	void saveStorage(std::string filename);

	virtual unsigned short in(unsigned char port) override;
	virtual void out(unsigned char port, unsigned short value) override;

private:
	unsigned char commandPort;
	unsigned char dataPort;

	unsigned short *storage;
	unsigned int storageSize;

	enum Command {
		None,
		Read,
		Write
	};

	Command currentCommand;
	int counter;
	unsigned short blockAddress;

	unsigned short operationBuffer[blockSize];
};

CCPUStorage::CCPUStorage(std::string imageFile, unsigned char commandPort, unsigned char dataPort)
 : commandPort(commandPort), dataPort(dataPort), storageSize(blockSize*numBlocks), currentCommand(None)
{
	std::ifstream image(imageFile.c_str(),std::ios::binary);

	image.seekg(0,std::ios::end);
	int imageSize=image.tellg();
	image.seekg(0,std::ios::beg);

	storage=new unsigned short[storageSize];
	memset(storage,0,storageSize*sizeof(unsigned short));
	image.read((char*)storage,imageSize);
	fixBufferEndianess(storage,storageSize);
}

void CCPUStorage::saveStorage(std::string filename)
{
	std::ofstream image(filename.c_str(),std::ios::binary);

	fixBufferEndianess(storage,storageSize);
	image.write((char*)storage,storageSize*sizeof(unsigned short));
	fixBufferEndianess(storage,storageSize);
}

unsigned short CCPUStorage::in(unsigned char port)
{
	if(port!=dataPort) {
		return 0;
	}

	unsigned short result;

	switch(currentCommand) {
	case None:
		break;
	case Read:
		result=operationBuffer[counter++];
		if(counter==blockSize) {
			currentCommand=None;
		}
		break;
	case Write:
		break;
	}

	return result;
}

void CCPUStorage::out(unsigned char port, unsigned short value)
{
	if(port==commandPort)  {
		blockAddress=value & 0x7FFF;

		if(value & 0x8000) {
			currentCommand=Write;
		} else {
			currentCommand=Read;
			memcpy(operationBuffer,storage+blockAddress*blockSize,blockSize*sizeof(unsigned short));
		}
		counter=0;
	} else if(port==dataPort) {
		if(currentCommand==Write) {
			operationBuffer[counter++]=value;
			if(counter==blockSize) {
				currentCommand=None;
				memcpy(storage+blockAddress*blockSize,operationBuffer,blockSize*sizeof(unsigned short));
			}
		}
	}
}

class CCPULinePrinter : public CCPUDevice {
public:
	CCPULinePrinter(unsigned char port) : port(port) {};

	virtual unsigned short in(unsigned char port) override { return 0; };
	virtual void out(unsigned char port, unsigned short value) override;

private:
	unsigned char port;
};

void CCPULinePrinter::out(unsigned char port, unsigned short value)
{
	if(port==this->port) {
		std::cout << (char)(value & 0xFF);
	}
}

class CCPUKeyboard : public CCPUDevice {
public:
	CCPUKeyboard(unsigned char port) : port(port), currentKey(0) {};

	virtual unsigned short in(unsigned char port) override { if(port==this->port) { return currentKey; } return 0; };
	virtual void out(unsigned char port, unsigned short value) override {};

	void setCurrentKey(unsigned short currentKey) { this->currentKey=currentKey; };

private:
	unsigned char port;
	unsigned short currentKey;
};

class CCPU {
public:
	CCPU();

	void registerDeviceForPort(unsigned char port, CCPUDevice *device);

	void reset();

	void step();

	unsigned short *getMemory() { return memory; };

	void dumpMemory(std::string filename);

private:
	unsigned short regA, regB, regC;
	unsigned short regIP, regSP, regFLAGS;
	unsigned short instructionStartIP; // required for instructions referencing IP

	unsigned short memory[65536];

	CCPUDevice *ports[256];

	unsigned short getOperand(CCPUOperandCode operand);
	void writeOperand(CCPUOperandCode operand, unsigned short value);
};

CCPU::CCPU()
{
	memset(ports,0,256*sizeof(void*));
	reset();
}

void CCPU::registerDeviceForPort(unsigned char port, CCPUDevice *device)
{
	ports[port]=device;
}

void CCPU::reset()
{
	regA=regB=regC=0;
	regIP=0;
	regFLAGS=0;
	regSP=0x8000;

	memset(memory,0,65536*sizeof(unsigned short));

	// boot loader machine code
	const unsigned short bootloader[] =
	{
		0xCAA0, 0x0000, 0x0000,			//     OUT 0, 0x0000
		0x7A00, 0x0000,					//     MOVE 0, A
										// loadbootcode:
		0xBAC0, 0x0001, 0x0100,			//     IN 1, [0x100+A]
		0x10A0, 0x0001,					//     ADD A, 1, A
		0x90A0, 0x0200,					//     TL A, 512
		0xF00A, 0x0005,					//     CJ loadbootcode (0x0005)
		0x500A, 0x0100					//     JMP 0x100
	};

	for(int i=0;i<16;i++) {
		memory[i]=bootloader[i];
	}
}

unsigned short CCPU::getOperand(CCPUOperandCode operand)
{
	unsigned short value;

	switch(operand) {
	case RegisterA:
		value=regA;
		break;
	case RegisterB:
		value=regB;
		break;
	case RegisterC:
		value=regC;
		break;
	case RegisterIP:
		value=instructionStartIP;
		break;
	case RegisterSP:
		value=regSP;
		break;
	case RegisterFLAGS:
		value=regFLAGS;
		break;
	case MemoryA:
		value=memory[regA];
		break;
	case MemoryB:
		value=memory[regB];
		break;
	case MemoryC:
		value=memory[regC];
		break;
	case MemorySP:
		value=memory[regSP];
		break;
	case Constant:
		value=memory[regIP++];
		break;
	case MemoryConstant:
		value=memory[memory[regIP++]];
		break;
	case MemoryConstantPlusA:
		value=memory[memory[regIP++]+regA];
		break;
	case MemoryConstantPlusIP:
		value=memory[memory[regIP++]+instructionStartIP]; // IP was already incremented in step(), so -1 here
		break;
	case MemoryConstantPlusSP:
		value=memory[memory[regIP++]+regSP];
		break;
	case MemoryBPlusA:
		value=memory[regB+regA];
		break;
	default:
		// other values for operand code are undefined
		value=regA;
		break;
	};

	return value;
}

void CCPU::writeOperand(CCPUOperandCode operand, unsigned short value)
{
	switch(operand) {
	case RegisterA:
		regA=value;
		break;
	case RegisterB:
		regB=value;
		break;
	case RegisterC:
		regC=value;
		break;
	case RegisterIP:
		regIP=value;
		break;
	case RegisterSP:
		regSP=value;
		break;
	case RegisterFLAGS:
		regFLAGS=value;
		break;
	case MemoryA:
		memory[regA]=value;
		break;
	case MemoryB:
		memory[regB]=value;
		break;
	case MemoryC:
		memory[regC]=value;
		break;
	case MemorySP:
		memory[regSP]=value;
		break;
	case Constant:
		memory[regIP++]=value;
		break;
	case MemoryConstant:
		memory[memory[regIP++]]=value;
		break;
	case MemoryConstantPlusA:
		memory[memory[regIP++]+regA]=value;
		break;
	case MemoryConstantPlusIP:
		memory[memory[regIP++]+instructionStartIP]=value; // IP was already incremented in step(), so -1 here
		break;
	case MemoryConstantPlusSP:
		memory[memory[regIP++]+regSP]=value;
		break;
	case MemoryBPlusA:
		memory[regB+regA]=value;
		break;
	default:
		// other values for operand code are undefined
		break;
	};
}

void CCPU::step()
{
	unsigned short instructionWord=memory[regIP];

	// decode instruction
	CCPUInstruction instruction=CCPUInstruction((instructionWord & 0xF000) >> 12);
	CCPUOperandCode source1=CCPUOperandCode((instructionWord & 0x0F00) >> 8);
	CCPUOperandCode source2=CCPUOperandCode((instructionWord & 0x00F0) >> 4);
	CCPUOperandCode target=CCPUOperandCode((instructionWord & 0x000F));

	instructionStartIP=regIP;
	regIP++;
//	std::cout << "\t" << regIP << std::endl;

	unsigned short source1Value, source2Value, targetValue;

	switch(instruction) {
	case InstNOP:
		break;
	case InstADD:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		writeOperand(target,source1Value+source2Value);
		if(source1Value+(int)source2Value>65536) {
			regFLAGS|=FlagOverflow;
		}
		break;
	case InstSUB:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		writeOperand(target,source1Value-source2Value);
		if(source1Value-(int)source2Value<0) {
			regFLAGS|=FlagOverflow;
		}
		break;
	case InstMUL:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		writeOperand(target,source1Value*source2Value);
		if(source1Value*(int)source2Value>65536) {
			regFLAGS|=FlagOverflow;
		}
		break;
	case InstDIV:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		if(source2Value!=0) {
			writeOperand(target,source1Value/source2Value);
		} else {
			regFLAGS|=FlagOverflow;
		}
		break;
	case InstJUMP:
		targetValue=getOperand(target);
		regIP=targetValue;
		break;
	case InstCALL:
		targetValue=getOperand(target);
		memory[--regSP]=regIP;
		regIP=targetValue;
		break;
	case InstMOVE:
		source1Value=getOperand(source1);
		writeOperand(target,source1Value);
		break;
	case InstTE:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		if(source1Value==source2Value) {
			regFLAGS|=FlagTest;
		} else {
			regFLAGS&=0xFFFF-FlagTest;
		}
		break;
	case InstTL:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		if(source1Value<source2Value) {
			regFLAGS|=FlagTest;
		} else {
			regFLAGS&=0xFFFF-FlagTest;
		}
		break;
	case InstTLE:
		source1Value=getOperand(source1);
		source2Value=getOperand(source2);
		if(source1Value<=source2Value) {
			regFLAGS|=FlagTest;
		} else {
			regFLAGS&=0xFFFF-FlagTest;
		}
		break;
	case InstIN:
		source1Value=getOperand(source1) % 0x100;
		if(ports[source1Value]!=0) {
			source2Value=ports[source1Value]->in(source1Value);
		}
		writeOperand(source2,source2Value);
		break;
	case InstOUT:
		source1Value=getOperand(source1) % 0x100;
		source2Value=getOperand(source2);
		if(ports[source1Value]!=0) {
			ports[source1Value]->out(source1Value,source2Value);
		}
		break;
	case InstPUSH:
		source1Value=getOperand(source1);
		memory[--regSP]=source1Value;
		break;
	case InstPOP:
		source1Value=memory[regSP++];
		if(regSP>0x8000) {
			regSP=0x8000;
		}
		writeOperand(target,source1Value);
		break;
	case InstCJ:
		targetValue=getOperand(target);
		if(regFLAGS & FlagTest) {
			regIP=targetValue;
		}
		break;
	};
}

void CCPU::dumpMemory(std::string filename)
{
	std::ofstream dump(filename.c_str(),std::ios::binary);

	fixBufferEndianess(memory,65536);
	dump.write((char*)memory,65536*sizeof(unsigned short));
	fixBufferEndianess(memory,65536);
}

int SDL_main(int argc, char *argv[])
{
	CCPUStorage storage("input.image",0,1);
	CCPULinePrinter linePrinter(2);
	CCPUKeyboard keyboard(3);

	CCPU ccpu;

	ccpu.registerDeviceForPort(0,&storage);
	ccpu.registerDeviceForPort(1,&storage);
	ccpu.registerDeviceForPort(2,&linePrinter);
	ccpu.registerDeviceForPort(3,&keyboard);

	SDL_Init(SDL_INIT_VIDEO);
	SDL_EnableUNICODE(1);

	SDL_Surface *screen=SDL_SetVideoMode(320,200,24,SDL_SWSURFACE);

	SDL_Event event;

	bool doContinue=true;
	while(doContinue) {
		for(int i=0;i<10000;i++) {
			ccpu.step();
		}

		SDL_LockSurface(screen);
		unsigned char *memory=reinterpret_cast<unsigned char*>(ccpu.getMemory())+65536;
		unsigned short *palette=ccpu.getMemory()+32768+32000;
		for(unsigned int y=0;y<200;y++) {
			for(unsigned int x=0;x<320;x++) {
				unsigned char colour=*(memory++);
				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel]=palette[3*colour+2];
				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel+1]=palette[3*colour+1];
				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel+2]=palette[3*colour];
//				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel]=0;
//				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel+1]=0;
//				reinterpret_cast<unsigned char*>(screen->pixels)[y*screen->pitch+x*screen->format->BytesPerPixel+2]=255;
			}
		}
		SDL_UnlockSurface(screen);
		SDL_UpdateRect(screen,0,0,0,0);

		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_MOUSEMOTION:
				break;
			case SDL_MOUSEBUTTONDOWN:
				break;
			case SDL_MOUSEBUTTONUP:
				break;
			case SDL_KEYDOWN:
				keyboard.setCurrentKey(event.key.keysym.unicode);
				break;
			case SDL_KEYUP:
				keyboard.setCurrentKey(0);
				break;
			case SDL_ACTIVEEVENT:
				break;
			case SDL_QUIT:
				doContinue=false;
				break;
			}
		}
	}

	SDL_Quit();

	storage.saveStorage("final.image");
	ccpu.dumpMemory("memory.dump");

	std::cin.get();
	
	return 0;
}
